Package: RNAmf
Type: Package
Title: Recursive Non-Additive Emulator for Multi-Fidelity Data
Version: 1.1.4
Authors@R: c(
	person(given="Junoh", family="Heo", role=c("aut", "cre"), email="heojunoh@msu.edu"), 	     	person(given="Chih-Li", family="Sung", role="aut", email="sungchih@msu.edu")
	)
Maintainer: Junoh Heo <heojunoh@msu.edu>
Description: Performs RNA emulation and active learning proposed by Heo and Sung (2025) <doi:10.1080/00401706.2024.2376173> for multi-fidelity computer experiments. The RNA emulator is particularly useful when the simulations with different fidelity level are nonlinearly correlated. The hyperparameters in the model are estimated by maximum likelihood estimation. 
License: GPL-3
Encoding: UTF-8
Imports: plgp, stats, methods, lhs, doParallel, foreach, doRNG, fields,
        mvtnorm, ggplot2, ggpubr, scales
Suggests: knitr, rmarkdown
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-05-05 02:51:06 UTC; junoh
Author: Junoh Heo [aut, cre],
  Chih-Li Sung [aut]
Repository: CRAN
Date/Publication: 2026-05-05 05:20:02 UTC
Built: R 4.6.0; ; 2026-05-05 05:50:22 UTC; unix
