# Generated by roxygen2: do not edit by hand

S3method(coef,RNAmf)
S3method(fitted,RNAmf)
S3method(plot,RNAmf)
S3method(predict,RNAmf)
S3method(print,RNAmf)
S3method(print,summary.RNAmf)
S3method(residuals,RNAmf)
S3method(summary,RNAmf)
export(AL_RNAmf)
export(NestedX)
export(RNAmf)
export(imputer_RNA)
import(ggplot2)
importFrom(doParallel,registerDoParallel)
importFrom(doParallel,stopImplicitCluster)
importFrom(doRNG,"%dorng%")
importFrom(fields,rdist)
importFrom(foreach,foreach)
importFrom(ggpubr,ggarrange)
importFrom(lhs,maximinLHS)
importFrom(lhs,randomLHS)
importFrom(methods,is)
importFrom(plgp,covar.sep)
importFrom(plgp,distance)
importFrom(scales,hue_pal)
importFrom(stats,fitted)
importFrom(stats,optim)
importFrom(stats,pnorm)
importFrom(stats,predict)
importFrom(stats,quantile)
importFrom(stats,rnorm)
importFrom(stats,setNames)
importFrom(stats,uniroot)
importFrom(stats,var)
